import numpy as np


class Wavefront:
    """A class representing a monochromatic wavefront. :class:`Wavefront` is
    used internally by Lentil to perform diffraction propagation calculations.

    Parameters
    ----------
    wavelength : float
        Wavelength in meters

    shape : array_like
        Wavefront shape

    pixelscale : float, optional
        Wavefront array spatial sampling in meters/pixel

    Attributes
    ----------
    focal_length : float or np.inf
        Wavefront focal length. A plane wave (default) has an infinite focal
        length (``np.inf``).

    tilt : list
        List of objects which implement a ``shift`` method. This method should
        accept the following parameters:

        ``shift(xs, ys, z, wavelength)``

        and return an updated x and y shift.

    """
    def __init__(self, wavelength, shape=None, pixelscale=None, planetype=None):

        self.wavelength = wavelength
        self.pixelscale = pixelscale
        self.planetype = planetype

        # All new Wavefront objects represent a perfect plane wave
        if shape:
            self.data = np.ones((1, shape[0], shape[1]), dtype=np.complex128)
        else:
            self.data = np.array([[1.]], dtype=np.complex128)

        # Wavefront focal length (which is infinity for a plane wave)
        self.focal_length = np.inf

        self.tilt = []  # List of pre-propagation tilts
        self.pixel_shift = []

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = np.asarray(value)

    @property
    def shape(self):
        """Wavefront array shape"""
        return self.data.shape

    @property
    def depth(self):
        """Number of individual Wavefront arrays in :attr:`data`"""
        return self.data.shape[0]

    @property
    def planetype(self):
        return self._planetype

    @planetype.setter
    def planetype(self, value):
        assert value in {'pupil', 'image', None}
        self._planetype = value

    def shift(self, pixelscale, oversample):
        """Compute image plane shift due to wavefront tilt.

        This is a somewhat tricky method. Fundamentally it iterates over the
        :attr:`~lentil.Wavefront.tilt` list and computes the resulting shift in
        terms of number of pixels in oversampled space. This calculation is
        complicated by the fact that in some cases, an element in
        :attr:`~lentil.Wavefront.tilt` will itself be a list. In this case, the
        shift should be tracked individually for each entry in the list. All
        ensuing calculations should be done in parallel (i.e. the
        multi-dimensional shift array should not be recollapsed. This behavior
        allows SegmentedPupil to handle segment tilts individually.

        Parameters
        ----------
        pixelscale : float
            Image plane spatial sampling in meters/pixel

        oversample : int
            Oversampling factor

        Returns
        -------
        shift : (depth, 2) ndarray
            Image plane shift in number of (possibly oversampled) pixels

        """

        # Example:
        # tilt = [Shift(10,10), [Shift(100,100), Shift(200,200)], Shift(50,50)]
        # Beginning shift = [0,0]
        # After first tilt:
        #   shift = [10,10]
        # After second tilt, shift is duplicated before each shift is applied:
        #   shift = [[110,110], [210,210]]
        # All successive tilts are now applied in parallel:
        #   shift = [[160,160], [260,260]]

        shift = np.zeros((1, 2))

        for tilt in self.tilt:
            if isinstance(tilt, list):

                # Reshape the shift array. It should have shape (len(tilt), 2).
                # If shift.shape is (1,2), we'll duplicate shift along the 0
                # axis so that it has shape (len(tilt),2). If shift.shape is
                # anything else, we can assume that the above duplication has
                # already happened so we'll just verify that the sizes have
                # remained consistent.
                if shift.shape[0] == 1:
                    shift = np.repeat(shift, len(tilt), axis=0)
                else:
                    assert shift.shape[0] == len(tilt)

                # Now we can iterate over the tilts
                for d, t in enumerate(tilt):
                    shift[d, 0], shift[d, 1] = t.shift(xs=shift[d, 0],
                                                       ys=shift[d, 1],
                                                       z=self.focal_length,
                                                       wavelength=self.wavelength)
            else:
                for d in np.arange(shift.shape[0]):
                    shift[d, 0], shift[d, 1] = tilt.shift(xs=shift[d, 0],
                                                          ys=shift[d, 1],
                                                          z=self.focal_length,
                                                          wavelength=self.wavelength)

        return (shift/pixelscale) * oversample




class Cache:
    """A dict-like object for caching data."""

    def __init__(self):
        self._cache = {}

    def set(self, key, value):
        """Set a cache value.

        Parameters
        ----------
        key : string

        value : object

        """
        self._cache[key] = value

    def get(self, key, default=None):
        """Get a cache value. If the object doesn't exist in the cache, the
        default value is returned.

        Parameters
        ----------
        key : string

        default : object, optional

        Returns
        -------
        value
            Cached value

        Note
        ----
        The literal value `None` should not be stored in a cache because it is
        impossible to distinguish between the stored value `None` and a cache
        miss signified by a returned value of `None`.

        """
        try:
            return self._cache[key]
        except KeyError:
            return default

    def add(self, key, value):
        """Set a cache value only if it doesn't already exist in the cache.

        Parameters
        ----------
        key : string

        value

        """
        if key not in self.keys():
            self._cache[key] = value

    def delete(self, key):
        """Delete a cache value

        Parameters
        ----------
        key : string

        """
        try:
            self._cache.pop(key)
        except KeyError:
            pass

    def clear(self):
        """Delete all the cache values."""
        self._cache = {}

    def keys(self):
        return self._cache.keys()



import numpy as np
from scipy.ndimage.interpolation import map_coordinates


def circle(shape, radius, center=(0, 0)):
    """Compute a circle with anti-aliasing.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    radius : float
        Radius of circle in pixels

    center : (2,) array_like, optional
        How far to shift center in float (rows, cols). Default is (0, 0).

    Returns
    -------
    circle : ndarray

    """
    x, y = mesh(shape)
    r = np.sqrt(np.square(x - center[1]) + np.square(y - center[0]))
    return np.clip(radius + 0.5 - r, 0.0, 1.0)


def circlemask(shape, radius, center=(0, 0)):
    """Compute a circular mask.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    radius : float
        Radius of circle in pixels

    center : array_like, optional
        How far to shift center in float (rows, cols). Default is (0, 0).

    Returns
    -------
    mask : ndarray

    """

    mask = circle(shape, radius, center)
    mask[mask > 0] = 1
    return mask


def hexagon(shape, radius, rotate=False):
    """Compute a hexagon mask.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    radius : int
        Radius of outscribing circle (which also equals the side length) in
        pixels.

    rotate : bool
        Rotate mask so that flat sides are aligned with the Y direction instead
        of the default orientation which is aligned with the X direction.

    Returns
    -------
    mask : ndarray

    """

    inner_radius = radius * np.sqrt(3)/2
    side_length = radius/2

    y, x = mesh(shape)

    rect = np.where((np.abs(x) <= side_length) & (np.abs(y) <= inner_radius))
    left_tri = np.where((x <= -side_length) & (x >= -radius) & (np.abs(y) <= (x + radius)*np.sqrt(3)))
    right_tri = np.where((x >= side_length) & (x <= radius) & (np.abs(y) <= (radius - x)*np.sqrt(3)))

    mask = np.zeros(shape)
    mask[rect] = 1
    mask[left_tri] = 1
    mask[right_tri] = 1

    if rotate:
        return mask.transpose()
    else:
        return mask


def slit(shape, width):
    """Compute a slit mask.

    Parameters
    ----------
    shape : array_like
        Size of output in pixels (nrows, ncols)

    width : int
        Slit width in pixels

    Returns
    -------
    mask : ndarray

    """

    y, x = mesh(shape)

    mask = np.zeros(shape)
    mask[np.abs(x) <= width/2] = 1

    return mask


def centroid(img):
    """Compute image centroid location.

    Parameters
    ----------
    img : array_like
        Input array.

    Returns
    -------
    tuple
        ``(x,y)`` centroid location.
    """

    img = np.asarray(img)
    img = img/np.sum(img)
    nr, nc = img.shape
    yy, xx = np.mgrid[0:nr, 0:nc]

    x = np.dot(xx.ravel(), img.ravel())
    y = np.dot(yy.ravel(), img.ravel())

    return x, y


def pad(array, shape):
    """Zero-pad an array.

    Note that pad works for both two and three dimensional arrays.

    Parameters
    ----------
    array : array_like
        Array to be padded.

    shape : tuple of ints
        Shape of output array in ``(nrows, ncols)``.

    Returns
    -------
    padded : ndarray
        Zero-padded array with shape ``(nrows, ncols)``. If ``array`` has a
        third dimension, the return shape will be ``(nrows, ncols, depth)``.
    """

    array = np.asarray(array)

    offset = 0
    if array.ndim == 3:
        offset = 1


    dr = shape[0] - array.shape[0+offset]
    dc = shape[1] - array.shape[1+offset]

    if dr <= 0:
        rmin0 = (array.shape[0+offset] - shape[0])//2
        rmax0 = rmin0 + shape[0]
        rmin1 = 0
        rmax1 = shape[0]
    else:
        rmin0 = 0
        rmax0 = array.shape[0+offset]
        rmin1 = (shape[0] - array.shape[0+offset])//2
        rmax1 = rmin1 + array.shape[0+offset]

    if dc <= 0:
        cmin0 = (array.shape[1+offset] - shape[1])//2
        cmax0 = cmin0 + shape[1]
        cmin1 = 0
        cmax1 = shape[1]
    else:
        cmin0 = 0
        cmax0 = array.shape[1]
        cmin1 = (shape[1] - array.shape[1+offset])//2
        cmax1 = cmin1 + array.shape[1+offset]

    if array.ndim < 3:
        padded = np.zeros((shape[0], shape[1]), dtype=array.dtype)
        padded[rmin1:rmax1, cmin1:cmax1] = array[rmin0:rmax0, cmin0:cmax0]
    else:
        padded = np.zeros((array.shape[0], shape[0], shape[1]), dtype=array.dtype)
        padded[:, rmin1:rmax1, cmin1:cmax1] = array[:, rmin0:rmax0, cmin0:cmax0]

    return padded


def boundary(img, threshold=0):
    """Find bounding row and column indices of data within an array."""
    img = np.asarray(img)
    img = (img > threshold)

    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)

    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    return rmin, rmax, cmin, cmax


def rebin(img, factor):
    """Rebin an image by an integer factor.

    Parameters
    ----------
    img : array_like
        Array or cube of arrays to rebin. If a cube is provided, the first dimension
        should index the image slices.

    factor : int
        Rebinning factor

    Returns
    -------
    img : ndarray
        Rebinned image

    See Also
    --------
    :func:`rescale`

    """
    img = np.asarray(img)

    if np.iscomplexobj(img):
        raise ValueError('rebin is not defined for complex data')

    if img.ndim == 3:
        rebinned_shape = (img.shape[0], img.shape[1]//factor, img.shape[2]//factor)
        img_rebinned = np.zeros(rebinned_shape, dtype=img.dtype)
        for i in range(img.shape[0]):
            img_rebinned[i] = img[i].reshape(rebinned_shape[1], factor,
                                             rebinned_shape[2], factor).sum(-1).sum(1)
    else:
        img_rebinned = img.reshape(img.shape[0]//factor, factor, img.shape[1]//factor,
                                   factor).sum(-1).sum(1)

    return img_rebinned


def gaussian2d(size, sigma):
    """2D Gaussian kernel."""
    x, y = np.meshgrid(np.linspace(-1, 1, size), np.linspace(-1, 1, size))

    G = np.exp(-((x**2/(2*sigma**2)) + (y**2/(2*sigma**2))))
    return G/np.sum(G)


def mesh(shape, shift=(0, 0)):
    """Generate a standard mesh."""

    nr = shape[0]
    nc = shape[1]

    x = np.arange(shape[1]) - np.floor(shape[1]/2.0) - shift[1]
    y = np.arange(shape[0]) - np.floor(shape[0]/2.0) - shift[0]

    return np.meshgrid(x, y)


def rescale(img, scale, shape=None, mask=None, order=3, mode='nearest',
            unitary=True):
    """Rescale an image by interpolation.

    Parameters
    ----------
    img : array_like
        Image to rescale

    scale : float
        Scaling factor. Scale factors less than 1 will shrink the image. Scale
        factors greater than 1 will grow the image.

    shape : array_like or int, optional
        Output shape. If None (default), the output shape will be the input img
        shape multiplied by the scale factor.

    mask : array_like, optional
        Binary mask applied after rescaling. If None (default), a mask is
        created from the nonzero portions of img. To skip masking operation,
        set ``mask = np.ones_like(img)``

    order : int, optional
        Order of spline interpolation used for rescaling operation. Default is
        3. Order must be in the range 0-5.

    mode : {'constant', 'nearest', 'reflect', 'wrap'}, optional
        Points outside the boundaries of the input are filled according to the
        given mode. Default is 'constant'.

    unitary : bool, optional
        Normalization flag. If True (default), a normalization is performed on
        the output such that the rescaling operation is unitary and image power
        (if complex) or intensity (if real) is conserved.

    Returns
    -------
    img : ndarray
        Rescaled image.

    Note
    ----
    The post-rescale masking operation should have no real effect on the
    resulting image but is included to eliminate interpolation artifacts that
    sometimes appear in large clusters of zeros in rescaled images.

    See Also
    --------
    :func:`rebin`

    """

    img = np.asarray(img)

    if mask is None:
        # take the real portion to ensure that even if img is complex, mask will
        # be real
        mask = np.zeros_like(img).real
        mask[img != 0] = 1

    if shape is None:
        shape = (int(img.shape[0]*scale), int(img.shape[1]*scale))
    else:
        if np.isscalar(shape):
            shape = (int(shape*scale), int(shape*scale))
        else:
            shape = (int(shape[0]*scale), int(shape[1]*scale))

    x = (np.arange(shape[1], dtype=np.float64) - shape[1]/2.)/scale + img.shape[1]/2.
    y = (np.arange(shape[0], dtype=np.float64) - shape[0]/2.)/scale + img.shape[0]/2.

    xx, yy = np.meshgrid(x, y)

    mask = map_coordinates(mask, [yy, xx], order=1, mode='nearest')
    mask[mask < np.finfo(mask.dtype).eps] = 0

    if np.iscomplexobj(img):
        out = np.zeros(shape, dtype=np.complex128)
        out.real = map_coordinates(img.real, [yy, xx], order=order, mode=mode)
        out.imag = map_coordinates(img.imag, [yy, xx], order=order, mode=mode)
    else:
        out = map_coordinates(img, [yy, xx], order=order, mode=mode)

    if unitary:
        out *= np.sum(img)/np.sum(out)

    out *= mask

    return out


def pixelscale_nyquist(wave, f_number):
    """Compute the output plane sampling which is Nyquist sampled for
    intensity.

    Parameters
    ----------
    wave : float
        Wavelength in meters

    f_number : float
        Optical system F/#

    Returns
    -------
    float
        Sampling in meters

    """
    return f_number * wave / 2


def make_index(mat):
    """Create a sparse coordinate list (COO) index dictionary.

    Parameters
    ----------
    mat : array_like
        Full matrix to ve vectorized

    Returns
    -------
    dict
        Index dictionary with the following attributes:

        * ``row`` List of row indices which contain nonzero data
        * ``col`` List of column indices which contain nonzero data
        * ``shape`` Tuple of dimensions of :attr:`~lentil.wfe.make_index.mat`

    See Also
    --------
    * :func:`~lentil.sparse.v2m` Convert sparse vectorized data to a full matrix
    * :func:`~lentil.sparse.m2v` Convert a full matrix to sparse vectorized data

    """
    mat = np.asarray(mat)
    row, col = mat.nonzero()
    shape = mat.shape
    return {'row': row, 'col': col, 'shape': shape}


def make_mask(index):
    """Create a mask from a sparse oordinate list (COO) index dictionary."""
    return v2m(np.ones(index['row'].shape), index)


def v2m(vec, index):
    """Convert a sparse vector to a full matrix.

    Parameters
    ----------
    vec : array_like
        Sparse vector to be reformed as a full matrix

    index : dict
        Corresponding index dictionary

    Returns
    -------
    ndarray
        Full matrix

    See Also
    --------
    * :func:`~lentil.sparse.m2v` Convert a full matrix to sparse vectorized data
    * :func:`~lentil.sparse.make_index` Create a sparse coordinate list (COO)
      index dictionary

    """
    mat = np.zeros(index['shape'])
    nnz = len(index['row'])
    for i in range(nnz):
        mat[index['row'][i], index['col'][i]] = vec[i]
    return mat


def m2v(mat, index):
    """Convert a full matrix to a sparse vector.

    Parameters
    ----------
    mat : array_like
        Full matrix to be reformed as a sparse vector

    index : dict
        Corresponding index dictionary

    Returns
    -------
    ndarray
        Sparse vector

    See Also
    --------
    * :func:`~lentil.sparse.v2m` Convert sparse vectorized data to a full matrix
    * :func:`~lentil.sparse.make_index` Create a sparse coordinate list (COO)
      index dictionary

    """
    raveled_indices = np.ravel_multi_index((index['row'], index['col']),
                                           index['shape'], order='C')
    vec = mat.ravel()
    return vec[raveled_indices]


def col_major_to_row_major(col_major_sens, col_major_index):
    """Convert a column-major sensitivity matrix and corresponding index
    dictionary (MATLAB-style) to a row-major sensitivity matrix and
    corresponding index disctionary (Python-style).

    Parameters
    ----------
    col_major_sens : array_like
        Sensitivity matrix ordered as column-major

    col_major_index : dict
        Corresponding index dictionary

    Returns
    -------
    ndarray
        Sensitivity matrix reordered as row-major
    dict
        Corresponding index dictionary

    Warning
    -------
    Keep in mind that when importing MATLAB-created index files to create
    :attr:`~lentil.wfe.col_major_to_row_major.col_major_index`, MATLAB used
    1-based indexing and Python uses 0-based indexing. As a result, both the
    row and col lists in
    :attr:`~lentil.wfe.col_major_to_row_major.col_major_index` must have one
    subtracted from them to ensure consistency. See example below.

    Examples
    --------
    .. code:: python3

        import numpy as np
        import lentil

        # import the column-major sensitivity matrix
        dwdx = np.genfromtxt('dwdx.csv', delimiter=',')

        # index_csv contains the row and column indices as its first two
        # columns. we'll have to manually make note of shape and directly
        # specify it
        index_csv = np.genfromtxt('index.csv', delimiter=',')

        index = {}
        index['row'] = np.asarray(index_csv[:,0], dtype=int)
        index['col'] = np.asarray(index_csv[:,1], dtype=int)
        index['shape'] = (255,255)

        # convert from 1-based (MATLAB) indexing to 0-based (Python) indexing
        index['row'] -= 1
        index['col'] -= 1

        # do the conversion
        dwdx, index = lentil.util.col_major_to_row_major(dwdx, index)

    See Also
    --------
    * :func:`~lentil.sparse.v2m` Convert sparse vectorized data to a full matrix
    * :func:`~lentil.sparse.m2v` Convert a full matrix to sparse vectorized data

    """

    col_major_sens = np.asarray(col_major_sens)

    # handle 1-dim vectors
    if col_major_sens.ndim == 1:
        col_major_sens = col_major_sens[..., np.newaxis]

    # make row_major_index
    mat = np.zeros(col_major_index['shape'])
    nnz = len(col_major_index['row'])
    for i in range(nnz):
        mat[col_major_index['row'][i], col_major_index['col'][i]] = 1
    row_major_index = make_index(mat)

    row_major_sens = np.zeros(col_major_sens.shape)

    for i in range(col_major_sens.shape[1]):
        m = v2m(col_major_sens[:, i], col_major_index)
        row_major_sens[:, i] = m2v(m, row_major_index)

    return row_major_sens, row_major_index


def omega(f_number):
    return np.pi*np.sin(np.arctan(1/(2*f_number)))**2



from operator import itemgetter

import numpy as np

from lentil import util
from lentil import fourier
from lentil.plane import Plane, Pupil, Image
from lentil.wavefront import Wavefront

__all__ = ['propagate']


def propagate(planes, wave, weight=None, npix=None, npix_chip=None, oversample=2,
              rebin=True, tilt='phase', interp_phasor=True, flatten=True,
              use_multiprocessing=False):
    """Compute a polychromatic point spread function using Fraunhofer
    diffraction.

    Parameters
    ----------
    planes : list_like
        List of :class:`~lentil.Plane` objects

    wave : array_like
        Array of wavelengths (in meters)

    weight : array_like, optional
        Weight multiple applied to each wavelength slice in :attr:`wave`.
        :attr:`weight` can be relative (for example, when considering the
        optical transmission through the optical system) or absolute (for
        example, when performing radiometrically accurate propagations where
        wavelength-dependent flux at the image plane is known). Must have
        the same length as :attr:`wave`. If not specified, ones are used.

    npix : int or (2,) tuple of ints, optional
        Shape of output plane. If not specified,
        ``npix = OpticalSystem.planes[-1].shape``.

    npix_chip : int or (2,) tuple of ints, optional
        Shape of propagation output plane. If None (default),
        ``npix_chip = npix``. If ``npix_chip != npix``, the propagation
        result is placed in the appropriate location in the output plane.
        npix_chip cannot be larger than npix.

    oversample : int, optional
        Number of times to oversample the output plane. Default is 2.

    rebin : bool, optional
        If ``True``, return the output plane in the sampling given by
        ``pixelscale``, binning down the output plane by a factor of
        ``oversample`` as needed. Note that this operation preserves power
        in the output plane. Default is ``True``.

    tilt : {'phase', 'angle'}, optional
        Propagation tilt handling strategy

        * 'phase' - any tilt present in the Element phase contribution is
          included in the Element*Wavefront product. (Default)
        * 'angle' - any tilt present in the Element phase is removed before
          computing the Element*Wavefront product. The equivalent angular
          tilt is included in the Wavefront's
          :attr:`~lentil.Wavefront.tilt` attribute.

    interp_phasor : bool, optional
        If True (default), the phasor components will be automatically
        interpolated to avoid aliasing and wraparound in the detector plane.
        If False, no checking or interpolation is performed.

    flatten : bool, optional
        If ``True``, the cube of wavelength-dependent output planes is
        flattened into a single 2D array before being returned. If
        ``False``, a cube of output planes is returned. Default is ``True``.

    Returns
    -------
    psf : ndarray
        Resulting point spread function.

    Note
    ----
    This function is a thin wrapper around Monocle's :class:`~lentil.prop.Propagate`
    object. If you need to do anything fancy that isn't provided  by the standard
    :func:`~lentil.propagate` method, consider subclassing or extending
    :class:`~lentil.prop.Propagate`.

    """

    with Propagate(planes) as p:
        psf = p.propagate(wave, weight, npix, npix_chip, oversample, rebin, tilt,
                          flatten)

    return psf


class Propagate:
    """Compute a polychromatic point spread function using Fraunhofer
    diffraction.

    Parameters
    ----------
    planes : list_like
        list of :class:`~lentil.Plane` objects

    Returns
    -------
    :class:`~lentil.prop.Propagate`

    Example
    -------
    ::

        with Propagate(planes) as p:
            psf = p.propagate(wave, weight, npix)

    """
    def __init__(self, planes):
        self.planes = planes

    def __enter__(self):
        # Pre-process and set up any static data we may need during the
        # propagation. Examples of this include setting caches, finding
        # dispersive elements, or really anything else that requires iterating
        # over the list of planes (which would be slow to perform with each
        # monochromatic wavefront propagation.)

        # Compute the wavefront shape required to support propagation through the
        # provided list of planes

        # The current approach is to find the maximum row and column dimensions of
        # only the Pupil elements in self.planes. This whole approach will have to
        # change once we support more complicated multi-plane propagations.

        shapes = [plane.shape for plane in self.planes if isinstance(plane, Pupil)]
        self.npix_wavefront = (max(shapes, key=itemgetter(0))[0],
                               max(shapes, key=itemgetter(1))[1])

        # Loop over the planes to build caches and identify any dispersive
        # elements
        for plane in self.planes:
            plane.cache_propagate()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Reset all the plane caches to their initial empty states
        for plane in self.planes:
            plane.clear_cache_propagate()

    def propagate(self, wave, weight=None, npix=None, npix_chip=None, oversample=2,
                  rebin=True, tilt='phase', flatten=True):
        """Compute a polychromatic point spread function using Fraunhofer
        diffraction.

        Parameters
        ----------
        wave : array_like
            Array of wavelengths (in meters)

        weight : array_like, optional
            Weight multiple applied to each wavelength slice in :attr:`wave`.
            :attr:`weight` can be relative (for example, when considering the
            optical transmission through the optical system) or absolute (for
            example, when performing radiometrically accurate propagations where
            wavelength-dependent flux at the image plane is known). Must have
            the same length as :attr:`wave`. If not specified, ones are used.

        npix : int or (2,) tuple of ints, optional
            Shape of output plane. If not specified,
            ``npix = OpticalSystem.planes[-1].shape``.

        npix_chip : int or (2,) tuple of ints, optional
            Shape of propagation output plane. If None (default),
            ``npix_chip = npix``. If ``npix_chip != npix``, the propagation
            result is placed in the appropriate location in the output plane.
            npix_chip cannot be larger than npix.

        oversample : int, optional
            Number of times to oversample the output plane. Default is 2.

        rebin : bool, optional
            If ``True``, return the output plane in the sampling given by
            ``pixelscale``, binning down the output plane by a factor of
            ``oversample`` as needed. Note that this operation preserves power
            in the output plane. Default is ``True``.

        tilt : {'phase', 'angle'}, optional
            Propagation tilt handling strategy

            * 'phase' - any tilt present in the Element phase contribution is
              included in the Element*Wavefront product. (Default)
            * 'angle' - any tilt present in the Element phase is removed before
              computing the Element*Wavefront product. The equivalent angular
              tilt is included in the Wavefront's
              :attr:`~lentil.Wavefront.tilt` attribute.

        interp_phasor : bool, optional
            If True (default), the phasor components will be automatically
            interpolated to avoid aliasing and wraparound in the detector plane.
            If False, no checking or interpolation is performed.

        flatten : bool, optional
            If ``True``, the cube of wavelength-dependent output planes is
            flattened into a single 2D array before being returned. If
            ``False``, a cube of output planes is returned. Default is ``True``.

        Returns
        -------
        psf : ndarray
            Resulting point spread function.

        """

        # TODO: move all this setup into __enter__:
        #   * npix
        #   * npix_chip
        #   * wave
        #   * weight
        if npix is None:
            npix = self.planes[-1].shape
        npix = np.asarray(npix)
        if npix.shape == ():
            npix = np.append(npix, npix)

        if npix_chip is None:
            npix_chip = npix
        npix_chip = np.asarray(npix_chip)
        if npix_chip.shape == ():
            npix_chip = np.append(npix_chip, npix_chip)

        wave = np.asarray(wave)
        if wave.shape == ():
            wave = wave[np.newaxis, ...]

        if weight is None:
            weight = np.ones(wave.shape)
        weight = np.asarray(weight)
        if weight.shape == ():
            weight = weight[np.newaxis, ...]

        # Create an empty output
        oversample_shape = (npix[0]*oversample, npix[1]*oversample)
        if flatten:
            output_shape = oversample_shape
        else:
            output_shape = (len(wave), oversample_shape[0], oversample_shape[1])

        output = np.zeros(output_shape)

        # We also need a temporary array to place the chips and compute intensity
        _output = np.zeros(oversample_shape, dtype=np.complex128)

        for n, (wl, wt) in enumerate(zip(wave, weight)):
            if wt > 0:
                w = Wavefront(wavelength=wl, pixelscale=None,
                              shape=self.npix_wavefront, planetype=None)

                w = self._propagate_mono(w, npix_chip, oversample, tilt)

                for d in range(w.depth):

                    # The shift term is given in terms of (x,y) but we place the chip in
                    # terms of (r,c)
                    # TODO: I think this will break if the last plane isn't a Detector
                    shift = np.flip(w.pixel_shift[d], axis=0)

                    # Compute the chip location
                    canvas_slice, chip_slice = _chip_insertion_slices(oversample_shape,
                                                                      (w.data.shape[1], w.data.shape[2]),
                                                                      shift)

                    # Insert the complex result in the output
                    if canvas_slice:
                        _output[canvas_slice] += w.data[d, chip_slice[0], chip_slice[1]]

                # Compute intensity
                if flatten:
                    output += np.abs(_output).real**2 * wt
                else:
                    output[n] = np.abs(_output).real**2 * wt

                # Zero out the local output array
                _output[:] = 0

        if rebin:
            output = util.rebin(output, oversample)

        return output

    def _propagate_mono(self, w, npix, oversample, tilt):
        """Propagate a monochromatic wavefront from plane to plane through the
        optical system using Fraunhofer diffraction.

        Parameters
        ----------
        w : :class:`~lentil.Wavefront`
            Wavefront object to propagate through the optical system.

        npix : int or (2,) ints
            Shape of output plane.

        oversample : int
            Number of times to oversample the output plane.

        tilt : {'phase', 'angle'}
            * 'phase' - any tilt present in the Element phase contribution is
              included in the Element*Wavefront product.
            * 'angle' - any tilt present in the Element phase is removed
              before computing the Element*Wavefront product. The equivalent
              angular tilt is included in the Wavefront's
              :attr:`~lentil.Wavefront.tilt` attribute.

        Returns
        -------
        field : ndarray
            Resulting complex field propagated though the optical system.

        """
        for plane, next_plane in _iterate_planes(self.planes):

            # Multiply by the current plane
            w = plane.multiply(w, tilt)

            # Now, we propagate to the next plane in the optical system

            # TODO: figure out how/when to accumulate tilt

            if (w.planetype == 'pupil') and isinstance(next_plane, Image):
                if next_plane.pixelscale is not None:
                    w = _propagate_pupil_image_fixed(w, next_plane.pixelscale, npix, oversample)
                else:
                    pass
            elif (w.planetype == 'image') and isinstance(next_plane, Pupil):
                pass
            elif (w.planetype == 'pupil') and isinstance(next_plane, Pupil):
                continue
            elif (w.planetype == 'image') and isinstance(next_plane, Image):
                continue
            elif isinstance(next_plane, Plane):
                continue
            else:
                raise TypeError('Unsupported propagation type ', w.planetype, ' to ', next_plane)

        return w


def _propagate_pupil_image_fixed(w, pixelscale, npix, oversample):

    # TODO: we should only apply the shift if this is the final plane
    shift = w.shift(pixelscale, oversample)
    w.tilt = []

    # Integer portion of the shift that will be accounted for
    # later
    w.pixel_shift = np.fix(shift)

    # Residual subpixel portion of the shift that is passed to
    # the DFT
    res_shift = shift - w.pixel_shift

    npix = npix * oversample

    alpha = (w.pixelscale * pixelscale) / (w.wavelength * w.focal_length * oversample)

    data = np.zeros((w.depth, npix[0], npix[1]), dtype=np.complex128)

    assert shift.shape[0] == w.depth, \
        'Dimension mismatch between tilt and wavefront depth'

    for d in range(w.depth):
        data[d] = fourier.dft2(w.data[d], alpha, npix, res_shift[d])

    w.data = data

    return w


class _iterate_planes:
    def __init__(self, planes):
        self.planes = planes
        self.length = len(planes)
        self.n = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.n < self.length-1:
            plane = self.planes[self.n]
            next_plane = self.planes[self.n+1]
            self.n += 1
            return plane, next_plane
        else:
            raise StopIteration()


def _chip_insertion_slices(npix_canvas, npix_chip, shift):
    npix_canvas = np.asarray(npix_canvas)
    npix_chip = np.asarray(npix_chip)

    # Canvas coordinates of the upper left corner of the shifted chip
    chip_shifted_ul = (npix_canvas / 2) - (npix_chip / 2) + shift

    # Chip slice indices
    chip_top = int(0)
    chip_bottom = int(npix_chip[0])
    chip_left = int(0)
    chip_right = int(npix_chip[1])

    # Canvas insertion slice indices
    canvas_top = int(chip_shifted_ul[0])
    canvas_bottom = int(chip_shifted_ul[0] + npix_chip[0])
    canvas_left = int(chip_shifted_ul[1])
    canvas_right = int(chip_shifted_ul[1] + npix_chip[1])

    # reconcile the chip and canvas insertion indices
    if canvas_top < 0:
        chip_top = -1 * canvas_top
        canvas_top = 0

    if canvas_bottom > npix_canvas[0]:
        chip_bottom -= canvas_bottom - npix_canvas[0]
        canvas_bottom = npix_canvas[0]

    if canvas_left < 0:
        chip_left = -1 * canvas_left
        canvas_left = 0

    if canvas_right > npix_canvas[1]:
        chip_right -= canvas_right - npix_canvas[1]
        canvas_right = npix_canvas[1]

    if np.any(np.array([canvas_bottom, chip_bottom, canvas_right, chip_right]) < 0):
        return None, None
    else:
        return (slice(canvas_top, canvas_bottom), slice(canvas_left, canvas_right)), \
               (slice(chip_top, chip_bottom), slice(chip_left, chip_right))



import numpy as np

__all__ = ['cached_property', 'normalize_power', 'iterable_amplitude', 'iterable_mask',
           'iterable_phase', 'iterable_segmask']


class cached_property:
    """Property that checks an object's cache for a cached value stored with the
    same attribute name before returning whatever value the in-place property
    defines.

    Note
    ----
    :class:`cached_property` is a drop-in replacement for Python's ``property``
    decorator. It has no side-effects and is safe to use even if Monocle's
    caching backend is not used.

    Example
    -------
    First we'll define a simple example class with a ``surface`` property
    using ``@cached_property``:

    .. code:: python3

        import lentil as le

        class CustomPlane(le.Plane):
            def __init__(self, surface):
                self._surface = surface

            @cached_property
            def surface(self):
                    return 10e-9

    Now we'll take a look at how a cached property behaves:

    .. code:: pycon

        >>> p = CustomPlane()
        >>> p.surface
        1e-08
        >>> p.cache.set('surface', 5e-6)
        >>> p.surface
        5e-06
        >>> p.cache.delete('surface')
        >>> p.surface
        1e-08

    """

    def __init__(self, fget=None, fset=None, fdel=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        self.__doc__ = getattr(fget, '__doc__')

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, cls=None):
        if instance.cache.get(self.name) is not None:
            return instance.cache.get(self.name)
        else:
            return self.fget(instance)

    def __set__(self, instance, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(instance, value)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel)


def normalize_power(array, power=1):
    r"""Normalizie the power in an array.

    The total power in an array is given by

    .. math::

        P = \sum{\left|\mbox{array}\right|^2}

    A normalization coefficient is computed as

    .. math::

        c = \sqrt{\frac{p}{\sum{\left|\mbox{array}\right|^2}}}

    The array returned by a will be scaled by the normalization coefficient so
    that its power is equal to :math:`p`.

    Parameters
    ----------
    array : array_like
        Array to be normalized

    power : float, optional
        Desired power in normalized array. Default is 1.

    Returns
    -------
    array : ndarray
        Normalized array

    """
    array = np.asarray(array)
    return array * np.sqrt(power/np.sum(np.abs(array)**2))


def iterable_amplitude(iterable):
    """Construct a common amplitude array from an iterable of planes

    Parameters
    ----------
    iterable : list_like
        List of planes

    Returns
    -------
    amplitude : ndarray

    """
    amplitude = np.array(1)
    for plane in iterable:
        amplitude = amplitude * plane.amplitude
    return amplitude


def iterable_phase(iterable):
    """Construct a common phase array from an iterable of planes

    Parameters
    ----------
    iterable: list_like
        List of planes

    Returns
    -------
    phase: ndarray

    """

    phase = np.array(0)
    for plane in iterable:
        phase = phase + plane.phase
    return phase


def iterable_mask(iterable):
    """Construct a common mask array from an iterable of planes

    Parameters
    ----------
    iterable: list_like
        List of planes

    Returns
    -------
    mask: ndarray

    """
    mask = np.array(1)
    for plane in iterable:
        mask = mask * plane.mask
    return mask


def iterable_segmask(iterable):
    """Construct a common segmask array from an iterable of planes

    Parameters
    ----------
    iterable: list_like
        List of planes

    Returns
    -------
    segmask: ndarray

    """
    if all(plane.segmask is None for plane in iterable):
        return None
    else:
        segmask = np.array(1)
        for plane in iterable:
            if plane.segmask is not None:
                segmask = segmask * plane.segmask
        return segmask



import numpy as np

__all__ = ['Pixel', 'Jitter', 'Smear']


class Convolvable:
    """Base class for representing spatial artifacts that can be applied via
    multiplication with a kernel in the frequency domain [1].

    See Also
    --------
    * :class:`~lentil.convolvable.Jitter`
    * :class:`~lentil.convolvable.Smear`
    * :class:`~lentil.convolvable.Pixel`

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Convolution_theorem

    """

    def __call__(self, *args, **kwargs):
        """Perform convolution

        This method defines the callable interface. Subclasses should provide a
        concrete implementation that computes an appropriate kernel and returns
        the result of :func:`convolve`.

        """
        raise NotImplementedError

    @staticmethod
    def convolve(f, G):
        r"""Convolve an array with a frequency domain kernel.

        Given an array :math:`f` in the spatial domain and a kernel :math:`G` in
        the frequency domain, the convolution is computed as

        .. math::

            f * g = \mathcal{F}^{-1} \left\{ \mathcal{F}\left\{f\right\}\cdot G \right\}


        Parameters
        ----------
        f : ndarray
            Spatial domain array

        G : ndarray
            Frequency domain kernel

        Returns
        -------
        out : ndarray
            f * g

        """
        return np.fft.ifft2(np.fft.fft2(f)*G)


class Pixel(Convolvable):
    """Callable object for applying the aperture effects of a square pixel on a
    discretely sampled irradiance distribution.

    Example
    -------
    Apply pixel MTF to a 3x oversampled PSF:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> pixel_mtf = lentil.convolvable.Pixel()
        >>> psf_mtf = pixel_mtf(psf, oversample=3)
        >>> psf_detector = lentil.util.rescale(psf_mtf, 1/3)

    Note that both the pixel MTF and detector resampling operations preserve
    radiometry:

    .. code:: pycon

        >>> print(np.sum(psf), np.sum(psf_mtf), np.sum(psf_detector))
        50398.80556524441 50398.80556524441 50398.80556524441

    """

    def __call__(self, img, oversample=1):
        """Convolve ``img`` with a square pixel transfer function.

        Parameters
        ----------
        img : array_like
            Input image

        oversample : int, optional
            Oversampling factor of img. Default is 1.

        Returns
        -------
        out : ndarray
            Image with pixel sampling effects applied.

        """
        img = np.asarray(img)
        x = np.fft.fftfreq(img.shape[1])
        y = np.fft.fftfreq(img.shape[0])
        kernel = self.kernel(x, y, oversample)

        return np.abs(self.convolve(img, kernel))

    def kernel(self, x, y, oversample):
        mtf_x = np.sinc(x*oversample)
        mtf_y = np.sinc(y*oversample)
        return np.dot(mtf_x[:, np.newaxis], mtf_y[np.newaxis, :])


class Jitter(Convolvable):
    """Callable object for applying image jitter via convolution.

    Parameters
    ----------
    pixelscale : float, optional
        Pixel size. If ``pixelscale = 1`` (default), the ``scale`` parameter
        provided when Jitter is called should give jitter in terms of fractional
        pixels. If ``pixelscale`` is the physical dimension of one pixel, ``the
        ``scale`` parameter provided when Jitter is called should give jitter in
        terms of the same units.

    Examples
    --------
    Apply 2 pixels jitter to a natively-sampled PSF:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> jitter = lentil.convolvable.Jitter()
        >>> psf_jitter = jitter(psf, scale=2)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_jitter)

    .. image:: ../_static/img/api/convolvable/jitter_px.png
        :width: 500px

    Apply 20 um jitter to a 3x oversampled PSF. Note that because we are
    specifying jitter in terms of linear distance on the focal plane, we must
    provide the detector pixelscale when creating the Jitter object. We also
    provide the oversampling factor, ensuring the convolution kernel is
    appropriately sized:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> jitter = lentil.convolvable.Jitter(pixelscale=5e-6)
        >>> psf_jitter = jitter(psf, scale=20e-6, oversample=3)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_jitter)

    .. image:: ../_static/img/api/convolvable/jitter_m.png
        :width: 500px

    """

    def __init__(self, pixelscale=1):
        self.pixelscale = pixelscale

    def __call__(self, img, scale, oversample=1):
        """Convolve ``img`` with a Gaussian jitter transfer function.

        Parameters
        ----------
        img : array_like
            Input image

        scale : float
            1-sigma jitter motion

            If ``pixelscale = 1``, ``scale`` should give jitter in terms of
            fractional pixels. If ``pixelscale`` is the physical dimension of
            one pixel, ``scale`` should give jitter in terms of the same units.

        oversample : int, optional
            Oversampling factor of img. Default is 1.

        Returns
        -------
        out : ndarray
            Image with jitter applied.

        """
        img = np.asarray(img)
        x = np.fft.fftfreq(img.shape[1])
        y = np.fft.fftfreq(img.shape[0])
        kernel = self.kernel(x, y, scale, oversample)
        out = np.abs(self.convolve(img, kernel))
        return out * np.sum(img)/np.sum(out)  # rescale to preserve input weight

    def kernel(self, x, y, scale, oversample):
        xx, yy = np.meshgrid(x, y)
        rho = np.sqrt(xx**2 + yy**2)
        return np.exp(-2*(np.pi*(scale/self.pixelscale)*oversample*rho)**2)


class Smear(Convolvable):
    """Callable object for applying image smear via convolution.

    Parameters
    ----------
    pixelscale : float, optional
        Pixel size. If ``pixelscale = 1`` (default), the ``scale`` parameter
        provided when Smear is called should give jitter in terms of fractional
        pixels. If ``pixelscale`` is the physical dimension of one pixel, ``the
        ``scale`` parameter provided when Smear is called should give jitter in
        terms of the same units.

    Examples
    --------
    Apply 100 um smear at 60 degrees to a natively-sampled PSF. Note that
    because we are specifying smear in terms of linear distance on the focal
    plane, we must provide the detector pixelscale when creating the Smear
    object:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> smear = lentil.convolvable.Smear(pixelscale=5e-6)
        >>> psf_smear = smear(psf, distance=100e-6, angle=60)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_smear)

    .. image:: ../_static/img/api/convolvable/smear_m.png
        :width: 500px

    Apply 10 pixels smear at a random angle to a 3x oversampled PSF:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> smear = lentil.convolvable.Smear()
        >>> psf_smear = smear(psf, distance=10, oversample=3)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_smear)

    .. image:: ../_static/img/api/convolvable/smear_px.png
        :width: 500px

    """

    def __init__(self, pixelscale=1):
        self.pixelscale = pixelscale

    def __call__(self, img, distance, angle=None, oversample=1):
        """Convolve ``img`` with a rotated sinc smear transfer function.

        Parameters
        ----------
        img : array_like
            Input image

        distance : float
            Linear smear distance

            If ``pixelscale = 1``, ``distance`` should give smear in terms of
            fractional pixels. If ``pixelscale`` is the physical dimension of
            one pixel, ``distance`` should give smear in terms of the same
            units.

        angle : float, optional
            Smear direction in degrees measured clockwise from the x-axis. If
            ``None`` (default), a new direction is randomly chosen every time
            :class:`Smear` is called.

        oversample : int, optional
            Oversampling factor of img. Default is 1.

        Returns
        -------
        out : ndarray
            Image with smear applied.

        """
        img = np.asarray(img)
        x = np.fft.fftfreq(img.shape[1])
        y = np.fft.fftfreq(img.shape[0])
        kernel = self.kernel(x, y, distance, angle, oversample)
        out = np.abs(self.convolve(img, kernel))
        return out * np.sum(img)/np.sum(out)  # rescale to preserve input weight

    def kernel(self, x, y, distance, angle, oversample):
        xx, yy = np.meshgrid(x, y)

        if angle is None:
            angle = np.random.uniform(0, 2*np.pi)
        else:
            angle = np.radians(angle)

        yy_rot = -np.sin(angle)*xx + np.cos(angle)*yy

        return np.sinc(yy_rot*(distance/self.pixelscale)*oversample)
